      subroutine psetup
c  Copyright (C) 2015, Jerry Sellwood
      implicit none
c Routine to setup initial particle positions for a run.  The preferred method
c   is from a file of initial coordinates, a .dfn file, but there is an option
c   to generate a specified disc configuration without a supplied file.  There
c   is no option supplied to generate random positions or velocities for a
c   non-disc population, however.
c For a quiet start, the number of particles generated is npr( icmp ) times
c   the number supplied in the .dfn file
c
c Radial coordinates for each population can be generated in three possible
c   ways as set by the logical variables smr and dist( icmp ) in / setup /
c   a) smr = T & dist = F - radii are chosen so as to give a smooth radial
c                           density profile for the given model
c   b) smr = F & dist = F - radii are generated randomly from a probability
c                           distribution (given by RPROB)
c   c) smr = F & dist = T - radii are read from a supplied .dfn file
c   (The combination smr = T & dist = T is nonsense and is trapped)
c
c Azimuthal coordinates are selected in one of two ways depending on the
c   logical variable quiet in / setup /
c   a) quiet = F - the azimuth of each particle is selected at random from a
c                  uniform distribution
c   b) quiet = T - a number npr particles are all placed at the same radius
c                  but spread evenly in azimuth - with a small random nudge
c                  set by the local variable amp0
c   The value of npr (the number of particles per ring) is set in this routine
c
c z-coordinates are generated randomly for disc particles from a Gaussian
c   distribution with a scale set by external ZTHICK
c
c Initial velocities are taken from the .dfn file when available, or are set
c   to the circular orbital velocity for a disc when no .dfn file is used.
c
c This routine would be inconvenient to parallize, especially when a quiet
c   start is used
c
c common blocks
c
      include 'inc/params.f'
c
      include 'inc/admin.f'
c
      include 'inc/buffer.f'
c
      include 'inc/grids.f'
c
      include 'inc/lunits.f'
c
      include 'inc/model.f'
c
c externals
      logical gtlogl, offgrd
      real ptrwgt, zthick
      real*8 gmassd, gmassh, rannor, ranuni, rprob, vcirc, zprob
c
c local array
      integer idim
      parameter ( idim = 500 )
      real pbuf( idim )
c
c local variables
      character*3 ext
      integer i, ifail, ind, irec, is, izon, jrec, k, norb
      integer np1, nrng, nrngs, nw, nwp, nx
      logical dfuqm, euler, rflsym, lh, luqmass, mdfs
      real a, amp0, cph, cps, cth, phfc, phbase, sph, sps, sth, vrad
      real vscale, vtan, vx, vy, vz, wp, wpm, wtp, x, xscale, y, z
      real*8 actmas, f, fr, f0, hz, phi, psi, rad, r1, r2, sigma
      real*8 theta, tol
      include 'inc/pi.f'
      save phi
c
      data vrad, vtan, vx, vy, vz, z / 6 * 0. /
c
c set new random seed
      call rvseed( 123456789 )
c initialize
      nx = 0
      do i = 1, nlists
        islist( 2, i, 1 ) = -1
      end do
c set flags that time step zones and time centering are not yet implemented
      lzones = .false.
      tcent = .false.
      lurand = .false.
c set defaults
      wp = 1
      dfuqm = .false.
c determine whether multiple DF files are expected
      k = 0
      do icmp = 1, ncmp
        if( dist( icmp ) )k = k + 1
      end do
      mdfs = k .gt. 1
c work over populations
      do icmp = 1, ncmp
        vrad = 0
        vtan = 0
        vx = 0
        vy = 0
        vz = 0
        z = 0
        wpm = 0
        if( nsp( icmp ) .gt. 0 )then
          if( master )then
            if( disc( icmp ) )then
              if( testp( icmp ) )then
                write( no, '( // 19x, ''Setting up test particles'' )' )
              else
                write( no, '( // 19x, ''Setting up disc particles'' )' )
              end if
            else
              write( no, '( // 19x, ''Setting up halo particles'' )' )
            end if
          end if
c look up which grid
          call switch( igrd( icmp ) )
          lh = .false.
          if( lheavy )then
c determine whether this component contains the heavy anchor particle
            do i = 1, ngrid
              if( jebind( i ) .eq. icmp )lh = .true.
            end do
          end if
c set npr, if not preset
          euler = .false.
          rflsym = .false.
          if( npr( icmp ) .eq. 0 )then
            if( quiet( icmp ) )then
              if( twod .or.
     +            sf3d )npr( icmp ) = ( 2 * ng - 1 ) / nsect + 1
              if( s3d .or. scf )npr( icmp ) = ( 2 * ng - 1 ) / nsect + 1
              if( c2d .or. c3d )npr( icmp ) = 10
              if( p3a )then
                rflsym = .true.
                npr( icmp ) = 2
              end if
              if( p3d )then
                npr( icmp ) = ( 2 * ng - 1 ) / nsect + 1
                rflsym = z0init( icmp ) .gt. 0.
                if( rflsym )npr( icmp ) = 4 * npr( icmp )
              end if
              if( npr( icmp ) .eq. 0 )call crash( 'PSETUP',
     +                                                   'npr not set' )
              euler = ( .not. disc( icmp ) ) .and.
     +                ( .not. sphrod( icmp ) )
              if( euler )npr( icmp ) = 3 * npr( icmp )
            else
              npr( icmp ) = 1
            end if
          end if
c update flag indicating whether particles will need to be randomized
          lurand = lurand .or. ( npr( icmp ) .gt. 1 ) .or. dist( icmp )
c check distribution function header, if requested
          if( dist( icmp ) )then
            if( smr( icmp ) )call crash( 'PSETUP',
     +                                     'Incompatible instructions' )
            luqmass = uqmass
c create extension name that may encode the component number
            ext = 'dfn'
            if( mdfs )write( ext( 3:3 ), '( i1 )' )icmp
            call opnfil( ndistf, ext, 'unformatted', 'old', 'seq', i )
            if( i .ne. 0 )then
              call crash( 'PSETUP', 'Failed to open file runN.'//ext )
            end if
            call dfhead( ndistf, irec, .false., .true. )
            if( uqmass .and. ( .not. luqmass ) )call
     +            crash( 'PSETUP', 'uqmass true required by .dfn file' )
c remember what the DF file contains and restore the global parameter
            dfuqm = uqmass
            uqmass = luqmass
            norb = 1
            do i = 1,3
              norb = norb * jdfcs( i, icmp )
            end do
            if( irec .gt. idim )call space(
     +                                    idim, irec, 'pbuf', 'PSETUP' )
            jrec = 0
            k = irec
            nwp = 3
            if( sphrod( icmp ) )nwp = 4
            if( dfuqm )nwp = nwp + 1
c allow for a possible extra heavy particle
            i = nsp( icmp )
            if( lheavy .and. lh )i = i - 1
            if( ( norb * npr( icmp ) ) .ne. i )then
              if( master )print
     +               '( ''Expected'', i8, ''/'', i2, '', found'', i8 )',
     +                                              i, npr( icmp ), norb
              call crash( 'PSETUP',
     +                      'Incorrect number of particles in DF file' )
            end if
            xscale = 1
            vscale = 1
            if( ( .not. cmprssd ) .and. ( ctype( icmp ) .ne. 'DFIT' )
     +   .and. ( .not. disc( icmp ) ) .and. ( nsp( icmp ) .gt. 0 ) .and.
     +          ( ( abs( cmpmas( icmp ) - 1. ) .gt. .001 ) .or.
     +            ( abs( rscale( icmp ) - 1. ) .gt. .001 ) ) )then
              if( gtlogl(
     +               'Do you want to scale the halo coordinates' ) )then
                xscale = rscale( icmp )
                vscale = sqrt( cmpmas( icmp ) / xscale )
              end if
            end if
          else
c inner and outer limits of active mass
            rhole = rinh( 1 ) / lscale
            r1 = rhole
            r2 = rtrunc( icmp )
c active mass fraction
            if( testp( icmp ) )then
              fmass( icmp ) = 1
            else
              if( disc( icmp ) )then
                f0 = gmassd( r1 )
                actmas = gmassd( r2 ) - f0
              else
                f0 = gmassh( r1 )
                actmas = gmassh( r2 ) - f0
              end if
c set fmass if not already set (eg from .pft file)
              if( fmass( icmp ) .eq. 0. )fmass( icmp ) =
     +                                           actmas / cmpmas( icmp )
            end if
c set up velocity dispersion if required
            if( .not. disc( icmp ) )then
              sigma = -1
              if( ( .not. dist( icmp ) ) .and.
     +            ( cdft( icmp ) .ne. 'JEAN' ) )then
c Let <V^2> = 3-d vely disperson of particles.  If c is the desired value
c of T/|W|, then the 1-D sigma^2 needed = 2/3 * c <V^2>
c for a uniform sphere <V^2> = 3GM/(5a) for vir eq
                if( ctype( icmp ) .eq. 'UNIS' )sigma =
     +                                sqrt( 2. * dfcns( 1, icmp ) / 5. )
c for an Aguilar-Merritt sphere <V^2> = 2GM/(3a) for vir eq
                if( ctype( icmp ) .eq. 'AGME' )then
                  sigma = sqrt( 4. * dfcns( 1, icmp ) / 9. )
                  if( ( dfcns( 1, icmp ) .gt. 0. ) .and.
     +                ( abs( dfcns( 3, icmp ) + 1 ) .gt. 1.e-6 )
     +          )call crash( 'PSETUP', 'sigma for halo not programmed' )
                end if
                if( sigma .lt. 0. )call crash( 'PSETUP',
     +                                        'sigma for halo not set' )
                if( master )write(
     +                       no, * )'1-D vel disp set to', sngl( sigma )
              end if
            end if
            if( master )then
              if( smr( icmp ) )then
                write( no, * )
     +                       ' Particles distributed smoothly in radius'
              else
                write( no, * )
     +               ' Radial coords chosen randomly from pby. distrbn.'
              end if
            end if
          end if
c set particle mass - default is by population
          if( uqmass .and. ( ncmp .gt. 1 ) )then
            wp = ptrwgt( icmp )
            if( wp .eq. 0. )then
              print *, icmp, cmpmas( icmp ), fmass( icmp )
              call crash( 'Psetup', 'massless particles!' )
            end if
            if( dfuqm )wtp = wp / fmass( icmp )
          end if
c check whether mesh is large enough for the particle distribution
          if( .not. ( noslfg .or. dr3d ) )then
            a = lscale * rtrunc( icmp )
            if( a .gt. rgrid( ngrid ) )print *,
     +'Warning from PSETUP: Outer radius', a, ' > rgrid', rgrid( ngrid )
c check coordinate type is consistent with grid
            if( threed ) then
              a = z0init( icmp ) * rscale( icmp ) * lscale
              if( a .gt. zm( jgrid ) )print *,
     +       'Warning from PSETUP: Thickness', a, ' > zmax', zm( jgrid )
            end if
          end if
c print if quiet start
          if( quiet( icmp ) )then
            if( master )write( no,
     +  '( i8, '' Particles placed evenly on each ring'' )' )npr( icmp )
c allow for a possible extra heavy particle
            i = nsp( icmp )
            if( lheavy .and. lh )i = i - 1
            if( mod( i, npr( icmp ) ) .ne. 0 )call crash( 'PSETUP',
     +                  'Number of particles is not a multiple of npr' )
          else
            if( master )write(
     +                no, * )' Azimuthal coordinates generated randomly'
          end if
          i = nsp( icmp )
          if( lh )i = i - 1
          nrngs = i / npr( icmp )
          amp0 = 0
          if( quiet( icmp ) )then
c choose amplitude of random nudge (the npr factor is for historical consistency)
            if( .not. testp( icmp ) )amp0 = .001 / real( npr( icmp ) )
c set angle between particles
            if( p3a )then
              phfc = 0
            else
              phfc = 2. * pi / real( nsect * npr( icmp ) )
            end if
c create a double ring of particles that can be reflection symmetric
            if( euler .or. rflsym )phfc = 2 * phfc
          end if
c
c generate particles
c
          is = 1
          do nrng = 1, nrngs
            if( dist( icmp ) )then
c get another particle
              k = k + nwp
              if( k .ge. irec )then
                nw = nwp * norb - jrec * irec
                nw = min( nw, irec )
                read( ndistf )( pbuf( i ), i = 1, nw )
                jrec = jrec + 1
                k = 0
              end if
              rad = xscale * pbuf( k + 1 )
              vrad = vscale * pbuf( k + 2 )
              vtan = vscale * pbuf( k + 3 )
              z = 0
              if( sphrod( icmp ) )z = xscale * pbuf( k + 4 )
              if( dfuqm )wp = wtp * pbuf( k + nwp )
              if( disc( icmp ) )then
                vz = 0
              else
                if( sphrod( icmp ) )then
c resolve vrad into components parallel to and perpendicular to plane
                  psi = pi * ( 1. - 2. * ranuni( 0. ) )
                  vz = vrad * sin( psi )
                  vrad = vrad * cos( psi )
c flip-half azimuthal velocities
                  x = ranuni( 0. )
                  if( x .lt. 0.5 )vtan = -vtan
                else
c choose angle for velocity projection
                  psi = pi * ( 1. - 2. * ranuni( 0. ) )
c restricted range of angles to eliminate retrograde particles
c                  psi = .5 * psi
                  if( .not. euler )then
                    cps = cos( psi )
                    sps = sin( psi )
                  end if
                end if
              end if
            else
              if( smr( icmp ) )then
c smooth radial distribution
                f = f0 + actmas * ( real( nrng ) - .5 ) / real( nrngs )
                r1 = 0.
                r2 = rtrunc( icmp )
                tol = .01 * rtrunc( icmp ) / real( nrngs )
                ind = 1
                ifail = 1
                do while ( ind .ne. 0 )
                  call fndzro( r1, r2, fr, tol, 0, pbuf, ind, ifail )
                  if( disc( icmp ) )then
                    fr = gmassd( r1 ) - f
                  else
                    fr = gmassh( r1 ) - f
                  end if
                end do
                rad = r1
                if( ifail .ne. 0 )then
                  print *, 'ifail = ', ifail
                  print *, f, r1, r2, gmassd( r1 ) - f, gmassd( r2 ) - f
                  call crash( 'PSETUP', 'Failed to find zero' )
                end if
              else if( testp( icmp ) )then
c equally spaced rings of test particles
                f = ( real( nrng ) - .5 ) / real( nrngs )
                rad = rhole + f * ( rtrunc( icmp ) - rhole )
              else
c generate random radii from the desired distribution
                rad = -1
                do while ( rad .lt. rhole )
                  rad = rtrunc( icmp )
                  rad = rprob( rad )
                end do
              end if
c velocities for a cold disc
              if( disc( icmp ) )then
                vrad = 0.
                vtan = vcirc( rad )
                vz = 0
              else
c generate isotropic velocities if dispersion is known
                if( sigma .gt. 0. )then
                  vx = rannor( 0.d0, sigma )
                  vy = rannor( 0.d0, sigma )
                  vz = rannor( 0.d0, sigma )
                end if
              end if
            end if
            if( twod )then
c convert coordinates to grid units
              a = rad * lscale
              vrad = vrad / gvfac
              vtan = vtan / gvfac
            else if( threed )then
c set initial z position
              if( disc( icmp ) )then
                hz = zthick( sngl( rad ) )
                if( hz .eq. 0. )then
                  z = 0
                else
                  a = zm( ngrid ) / lscale
                  z = 1.1 * a
                  do while ( abs( z ) .gt. a )
c random generator
                    z = zprob( hz )
                  end do
                end if
              else
c choose polar angle for z position
                theta = 1. - 2. * ranuni( 0. )
                if( euler )then
                  theta = acos( theta )
                else
                  cth = theta
                  sth = sqrt( 1. - cth * cth )
                end if
              end if
c convert coordinates to grid units
              a = rad * lscale
              z = z * lscale
              vrad = vrad / gvfac
              vtan = vtan / gvfac
              vx = vx / gvfac
              vy = vy / gvfac
              vz = vz / gvfac
            end if
c distribute particles around ring
            phbase = 2. * pi * ranuni( 0. )
            do i = 1, npr( icmp )
              phi = phbase + phfc * real( i - 1 )
              if( .not. euler )then
c random nudge
                if( quiet( icmp ) )phi = phi + amp0 * ranuni( 0. )
c confine to 0<phi<2pi
                phi = mod( phi, 2. * pi )
c place second half of particles symmetrically on the other side of disk plane
                if( rflsym .and. ( i .eq. 1 + npr( icmp ) / 2 ) )then
                  z = -z
                  vz = -vz
                  psi = psi + pi
                  cps = cos( psi )
                  sps = sin( psi )
                  cth = -cth
                end if
              else
c ditto for halo particles
                if( i .eq. 1 + npr( icmp ) / 2 )then
                  psi = psi + pi
                  cps = cos( psi )
                  sps = sin( psi )
                  cth = -cth
                end if
              end if
c check space
              if( is .gt. mbuff )then
                call relabl( mbuff )
                if( nx .gt. lpf )call crash( 'PSETUP',
     +                                        'Particle file overflow' )
                call scattr( mbuff )
                is = 1
              end if
c store coordinates
              if( twod )then
                x = a * cos( phi )
                y = a * sin( phi )
                newc( 1, is ) = x
                newc( 2, is ) = y
                newc( 3, is ) = ( x * vrad - y * vtan ) / a
                newc( 4, is ) = ( x * vtan + y * vrad ) / a
              else
                sph = sin( phi )
                cph = cos( phi )
                if( disc( icmp ) .or. sphrod( icmp ) )then
                  x = a * cph
                  y = a * sph
                  newc( 1, is ) = x
                  newc( 2, is ) = y
                  newc( 3, is ) = z
                  newc( 4, is ) = ( x * vrad - y * vtan ) / a
                  newc( 5, is ) = ( x * vtan + y * vrad ) / a
                  newc( 6, is ) = vz
                else
c coordinate transformation for spherical models
                  if( euler )then
                    newc( 1, is ) = a
                    newc( 2, is ) = 0
                    newc( 3, is ) = 0
                    newc( 4, is ) = vrad
                    newc( 5, is ) = vtan
                    newc( 6, is ) = 0
                    call rotate( psi, theta, phi, newc( 1, is ) )
                  else
                    newc( 1, is ) = a * sth * cph
                    newc( 2, is ) = a * sth * sph
                    newc( 3, is ) = a * cth
                    if( dist( icmp ) )then
                      newc( 4, is ) = vrad * sth * cph
     +                     - vtan * ( cps * sph + sps * cth * cph )
                      newc( 5, is ) = vrad * sth * sph
     +                     + vtan * ( cps * cph - sps * cth * sph )
                      newc( 6, is ) = vrad * cth + vtan * sps * sth
                    else
                      newc( 4, is ) = vx
                      newc( 5, is ) = vy
                      newc( 6, is ) = vz
                    end if
                  end if
                end if
              end if
c put all on-grid particles in zone 1 - zones are not useful in mkpcs
              izon = 1
c check for particles outside the grid
              if( offgrd( is ) )then
                izon = nlists
                noff = noff + 1
              end if
c assign labels
              iz( is ) = izon
              loc( is ) = nx
              iflag( is ) = icmp
              pwt( is ) = wp
              wpm = max( wpm, wp )
              label( is ) = igrd( icmp )
c increment counters
              nx = nx + nwpp
              is = is + 1
            end do
          end do
          if( dist( icmp ) )close( ndistf )
c add heavy anchor particle if required
          if( lheavy .and. lh )then
            if( is .gt. mbuff )then
              call relabl( mbuff )
              if( nx .gt. lpf )call crash( 'PSETUP',
     +                                        'Particle file overflow' )
              call scattr( mbuff )
              is = 1
            end if
c particle should be at rest in the center
            do i = 1, ndimen
              newc( i, is ) = 0
              newc( i + ndimen, is ) = 0
            end do
            iz( is ) = izon
            loc( is ) = nx
            iflag( is ) = icmp
c make it 10 times more massive than any other in this population
            pwt( is ) = 10. * wpm
            label( is ) = igrd( icmp )
c on master node for now
            jebind( jgrid ) = 0
            loch( jgrid ) = nx
c increment counters
            nx = nx + nwpp
            is = is + 1
          end if
c clear particles remaining in  / buffer /
          is = is - 1
          call relabl( is )
          if( nx .gt. lpf )call crash( 'PSETUP',
     +                                       'Particle array overflow' )
          call scattr( is )
c end main loop over populations
          if( .not. testp( icmp ) .and. master )write( no,
     +      '( '' Active mass fraction in cmp'', i2, '' is'', f10.6 )' )
     +                                               icmp, fmass( icmp )
        end if
      end do
c note number of particles outside grid, if any
      if( noff .gt. 0 .and. master )then
        write( no, * )noff, ' particles outside the grid at the start'
        print *, noff, ' particles outside the grid at the start'
      end if
c update list origin table
      do i = 1, nlists
        islist( 1, i, 1 ) = islist( 2, i, 1 )
      end do
c rescale mass of each particle so density is correct
c   ==> if this is not done with individual particle masses
      if( .not. uqmass )then
        np1 = 0
        do i = 1, ncmp
          if( ( .not. rigidp( i ) ) .and. ( .not. testp( i ) ) )then
            np1 = np1 + 1
c remember the component number that has massive particles
            icmp = i
          end if
        end do
        if( np1 .gt. 2 )then
          call crash( 'PSETUP',
     +            'pmass adjustment required for multiple populations' )
        else if( np1 .eq. 2 )then
          x = abs( fmass( 1 ) - fmass( 2 ) )
          if( x .gt. 1.d-3 .and. master )write( no, * )
     +    'fmass differs for two populations by', 100. * x / fmass( 1 ),
     +                   ' percent, values are:', fmass( 1 ), fmass( 2 )
          if( x .gt. 1.e-2 )call crash( 'PSETUP',
     +            'pmass adjustment required for multiple populations' )
        end if
        pmass = pmass * fmass( icmp )
        call slfset
      end if
c make sure icmp is undefined
      icmp = 0
      return
      end
